#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

static int trypatch20(int fd, off_t offset, char* name) {
	unsigned char buf[3];

	if (lseek(fd, offset, SEEK_SET) != offset) {
		return 20;
	}
	if (read(fd, buf, 3) != 3) {
		return 21;
	}
	if (!memcmp(buf, "\x03\x72\x20", 3)) {
		printf("%s, already patched\n", name);
		return 0;
	}
	if (memcmp(buf, "\x03\x75\x20", 3)) {
		return 10;
	}
	printf("%s, ", name);
	fflush(stdout);
	if (lseek(fd, offset + 1, SEEK_SET) != offset + 1) {
		printf("failed\n");
		return 30;
	}
	buf[0] = 0x72;
	if (write(fd, buf, 1) != 1) {
		printf("failed\n");
		return 40;
	}
	if (lseek(fd, offset, SEEK_SET) != offset) {
		printf("failed\n");
		return 41;
	}
	if (read(fd, buf, 3) != 3) {
		printf("failed\n");
		return 42;
	}
	if (!memcmp(buf, "\x03\x72\x20", 3)) {
		printf("now patched\n");
		return 0;
	}
	if (memcmp(buf, "\x03\x75\x20", 3)) {
		printf("failed\n");
		return 43;
	}
	printf("failed\n");
	return 44;
}

static int trypatch11(int fd, off_t offset, char* name) {
	unsigned char buf[3];

	if (lseek(fd, offset, SEEK_SET) != offset) {
		return 20;
	}
	if (read(fd, buf, 3) != 3) {
		return 21;
	}
	if (!memcmp(buf, "\x03\x72\x23", 3)) {
		printf("%s, already patched\n", name);
		return 0;
	}
	if (memcmp(buf, "\x03\x75\x23", 3)) {
		return 10;
	}
	printf("%s, ", name);
	fflush(stdout);
	if (lseek(fd, offset + 1, SEEK_SET) != offset + 1) {
		printf("failed\n");
		return 30;
	}
	buf[0] = 0x72;
	if (write(fd, buf, 1) != 1) {
		printf("failed\n");
		return 40;
	}
	if (lseek(fd, offset, SEEK_SET) != offset) {
		printf("failed\n");
		return 41;
	}
	if (read(fd, buf, 3) != 3) {
		printf("failed\n");
		return 42;
	}
	if (!memcmp(buf, "\x03\x72\x23", 3)) {
		printf("now patched\n");
		return 0;
	}
	if (memcmp(buf, "\x03\x75\x23", 3)) {
		printf("failed\n");
		return 43;
	}
	printf("failed\n");
	return 44;
}

static int nopatch(int fd, off_t offset, char* name) {
	unsigned char buf[3];

	if (lseek(fd, offset, SEEK_SET) != offset) {
		return 20;
	}
	if (read(fd, buf, 3) != 3) {
		return 21;
	}
	if (!memcmp(buf, "\x03\x75\x20", 3)) {
		printf("%s, does not need patch\n", name);
		return 0;
	}
	return 10;
}

static struct vmp {
	off_t offs;
	char* name;
	int (*fn)(int fd, off_t offset, char* name);
} vmpinfo[] = {
	{ 564792,  "VMware-476", trypatch20 },
	{ 0x9D527, "VMware-546", trypatch20 },
	{ 0x42BE5, "VMware-364", trypatch11 },
	{ 0x8E9FC, "VMware-570", nopatch },
	{ 0,       NULL,         NULL }};

int main(int argc, char* argv[]) {
	int fd;
	int err;
	struct vmp *vp;
		
	if (argc < 2) {
		fprintf(stderr, "File name missing!\n");
		return 99;
	}
	printf("Updating %s ... ", argv[1]);
	fflush(stdout);
	fd = open(argv[1], O_RDWR);
	if (fd == -1) {
		printf("failed\n");
		fprintf(stderr, "Cannot open %s: %m\n", argv[1]);
		return 98;
	}
	for (vp = vmpinfo; vp->name; vp++) {
		err = vp->fn(fd, vp->offs, vp->name);
		switch (err) {
			case 0:	close(fd); return 0;	/* OK */
			case 40 ... 49: 
				close(fd);
				fprintf(stderr, "ERROR! It is possible that your vmware binary is corrupted.\n");
				fprintf(stderr, "ERROR! Please, reinstall vmware.\n");
				return 40;
			case 30 ... 39:
				close(fd);
				fprintf(stderr, "ERROR! Cannot update vmware binary.\n");
				return 30;
			case 1 ... 29:
				break;
			default:
				close(fd);
				fprintf(stderr, "ERROR! Unexpected error.\n");
				return 99;
		}
	}
	close(fd);
	printf("Unknown version\n");
	fprintf(stderr, "ERROR! Sorry, unknown version of vmware is installed on this machine.\n");
	return 10;
}

